from typing import Any, Optional, List
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_handle, validate_did, validate_datetime


class Handle(BaseModel):
    """
    [none provided by spec]
    """

    seq: int = Field(...)
    did: str = Field(..., pre=True, validator=validate_did)
    handle: str = Field(..., pre=True, validator=validate_handle)
    time: str = Field(..., pre=True, validator=validate_datetime)


class Tombstone(BaseModel):
    """
    [none provided by spec]
    """

    seq: int = Field(...)
    did: str = Field(..., pre=True, validator=validate_did)
    time: str = Field(..., pre=True, validator=validate_datetime)


class Migrate(BaseModel):
    """
    [none provided by spec]
    """

    seq: int = Field(...)
    did: str = Field(..., pre=True, validator=validate_did)
    migrateTo: str = Field(...)
    time: str = Field(..., pre=True, validator=validate_datetime)


class Info(BaseModel):
    """
    [none provided by spec]
    """

    name: str = Field(..., known_values=["OutdatedCursor"])
    message: Optional[str] = Field(default=None)


class RepoOp(BaseModel):
    """
    [none provided by spec]
    """

    action: str = Field(..., known_values=["create", "update", "delete"])
    path: str = Field(...)
    cid: Any


class Commit(BaseModel):
    """
    [none provided by spec]
    """

    seq: int = Field(...)
    rebase: bool = Field(...)
    tooBig: bool = Field(...)
    repo: str = Field(..., pre=True, validator=validate_did)
    commit: Any
    prev: Any
    blocks: Any
    ops: Any
    blobs: List[Any] = Field(...)
    time: str = Field(..., pre=True, validator=validate_datetime)



from typing import Any
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_did, validate_datetime


class InviteCodeUse(BaseModel):
    """
    [none provided by spec]
    """

    usedBy: str = Field(..., pre=True, validator=validate_did)
    usedAt: str = Field(..., pre=True, validator=validate_datetime)


class InviteCode(BaseModel):
    """
    [none provided by spec]
    """

    code: str = Field(...)
    available: int = Field(...)
    disabled: bool = Field(...)
    forAccount: str = Field(...)
    createdBy: str = Field(...)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)
    uses: Any



from typing import Any, Optional, List
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import (
    validate_cid,
    validate_did,
    validate_datetime,
    validate_array,
    validate_at_uri,
    validate_handle,
)
from enum import auto, Enum
from psychonaut.api.lexicons.com.atproto.moderation.defs import ReasonType
from psychonaut.api.lexicons.com.atproto.server.defs import InviteCode
from psychonaut.api.lexicons.com.atproto.label.defs import Label


class RepoRef(BaseModel):
    """
    [none provided by spec]
    """

    did: str = Field(..., pre=True, validator=validate_did)


class ActionReversal(BaseModel):
    """
    [none provided by spec]
    """

    reason: str = Field(...)
    createdBy: str = Field(..., pre=True, validator=validate_did)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)


class VideoDetails(BaseModel):
    """
    [none provided by spec]
    """

    width: int = Field(...)
    height: int = Field(...)
    length: int = Field(...)


class ImageDetails(BaseModel):
    """
    [none provided by spec]
    """

    width: int = Field(...)
    height: int = Field(...)


class ActionType(str, Enum):
    TAKEDOWN = "takedown"
    FLAG = "flag"
    ACKNOWLEDGE = "acknowledge"


class ActionView(BaseModel):
    """
    [none provided by spec]
    """

    id: int = Field(...)
    action: ActionType
    subject: Any
    subjectBlobCids: List[str] = Field(...)
    createLabelVals: Optional[List[str]] = Field(default=None)
    negateLabelVals: Optional[List[str]] = Field(default=None)
    reason: str = Field(...)
    createdBy: str = Field(..., pre=True, validator=validate_did)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)
    reversal: Optional[ActionReversal] = None
    resolvedReportIds: List[int] = Field(...)


class ActionViewCurrent(BaseModel):
    """
    [none provided by spec]
    """

    id: int = Field(...)
    action: ActionType


class ReportView(BaseModel):
    """
    [none provided by spec]
    """

    id: int = Field(...)
    reasonType: ReasonType
    reason: Optional[str] = Field(default=None)
    subject: Any
    reportedBy: str = Field(..., pre=True, validator=validate_did)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)
    resolvedByActionIds: List[int] = Field(...)


class Moderation(BaseModel):
    """
    [none provided by spec]
    """

    currentAction: Optional[ActionViewCurrent] = None


class ModerationDetail(BaseModel):
    """
    [none provided by spec]
    """

    currentAction: Optional[ActionViewCurrent] = None
    actions: Any
    reports: Any


class RepoView(BaseModel):
    """
    [none provided by spec]
    """

    did: str = Field(..., pre=True, validator=validate_did)
    handle: str = Field(..., pre=True, validator=validate_handle)
    email: Optional[str] = Field(default=None)
    relatedRecords: List[Any] = Field(...)
    indexedAt: str = Field(..., pre=True, validator=validate_datetime)
    moderation: Moderation
    invitedBy: Optional[InviteCode] = None


class BlobView(BaseModel):
    """
    [none provided by spec]
    """

    cid: str = Field(..., pre=True, validator=validate_cid)
    mimeType: str = Field(...)
    size: int = Field(...)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)
    details: Optional[Any] = None
    moderation: Optional[Moderation] = None


class RepoViewDetail(BaseModel):
    """
    [none provided by spec]
    """

    did: str = Field(..., pre=True, validator=validate_did)
    handle: str = Field(..., pre=True, validator=validate_handle)
    email: Optional[str] = Field(default=None)
    relatedRecords: List[Any] = Field(...)
    indexedAt: str = Field(..., pre=True, validator=validate_datetime)
    moderation: ModerationDetail
    labels: Optional[Any] = None
    invitedBy: Optional[InviteCode] = None
    invites: Optional[Any] = None


class RecordView(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: str = Field(..., pre=True, validator=validate_cid)
    value: Any
    blobCids: List[str] = Field(..., pre=True, validator=validate_array(validate_cid))
    indexedAt: str = Field(..., pre=True, validator=validate_datetime)
    moderation: Moderation
    repo: RepoView


class RecordViewDetail(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: str = Field(..., pre=True, validator=validate_cid)
    value: Any
    blobs: Any
    labels: Optional[Any] = None
    indexedAt: str = Field(..., pre=True, validator=validate_datetime)
    moderation: ModerationDetail
    repo: RepoView


class ActionViewDetail(BaseModel):
    """
    [none provided by spec]
    """

    id: int = Field(...)
    action: ActionType
    subject: Any
    subjectBlobs: Any
    createLabelVals: Optional[List[str]] = Field(default=None)
    negateLabelVals: Optional[List[str]] = Field(default=None)
    reason: str = Field(...)
    createdBy: str = Field(..., pre=True, validator=validate_did)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)
    reversal: Optional[ActionReversal] = None
    resolvedReports: Any


class ReportViewDetail(BaseModel):
    """
    [none provided by spec]
    """

    id: int = Field(...)
    reasonType: ReasonType
    reason: Optional[str] = Field(default=None)
    subject: Any
    reportedBy: str = Field(..., pre=True, validator=validate_did)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)
    resolvedByActions: Any



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class UploadBlobResp(BaseModel):
    blob: Any


class UploadBlobReq(BaseModel):
    """
    Upload a new blob to be added to repo in a later request.
    """

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.repo.uploadBlob"

    async def do_xrpc(self, sess: Session) -> UploadBlobResp:
        resp = await sess.procedure(self)
        return UploadBlobResp(**resp)



import asyncio
from functools import wraps
import sys


def as_async(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))

    return wrapper



def print_error_and_fail(msg: str):
    print(msg, file=sys.stderr)
    sys.exit(1)


HANDLE_WARNING = """WARNING:

You input handle,
    {}
which looks like an email address but should probably look like,
    @{}
"""

def clean_handle(handle: str) -> str:
    if handle.startswith("@"):
        handle = handle[1:]

    idx = handle.find("@")
    if idx != -1 and (0 < idx < len(handle) - 1):
        suggestion = handle.replace("@", ".")
        print(HANDLE_WARNING.format(handle, suggestion), file=sys.stderr)
    

    return handle


from typing import Any, Dict, List, Protocol, Union
from psychonaut.lexicon.constants import MODULE_PREFIX

from psychonaut.lexicon.ctx import GenCtx
from psychonaut.lexicon.util import camel_to_class_name, camel_to_snake
from .types import (
    LexArray,
    LexInteger,
    LexObject,
    LexString,
    LexBoolean,
    LexXrpcParameters,
)


class FieldDef(Protocol):
    name: str
    type: str
    required: bool
    opts: Dict[str, Any]


def generate_pydantic_fields(
    ctx: GenCtx, params: Union[LexXrpcParameters, LexObject]
) -> List[str]:
    # Needs:
    # .required
    # .properties
    #
    # compiles: LexObject
    # compiles: LexXrpcParameters
    if not params or not params.properties:
        return []

    required_fields = set(params.required or [])

    return [
        _generate_field(ctx, prop_name, prop_name in required_fields, prop_type)
        for prop_name, prop_type in params.properties.items()
    ]


_lex_type_to_pydantic_type = {
    "string": "str",
    "integer": "int",
    "boolean": "bool",
}


def _consistent_opts_str(opts: Dict[str, Any]) -> str:
    alphabetical_opts = sorted(opts.items(), key=lambda x: x[0])
    opts_str = ", ".join(f"{k}={repr(v)}" for k, v in alphabetical_opts)
    return opts_str


def _build_expanded(
    pytype: str, required: bool, opts: Dict[str, Any], name: str
) -> str:
    src = ""

    if not required and "default" not in opts:
        opts["default"] = None

    opts_str = _consistent_opts_str(opts)

    opts_params = ""
    if opts_str:
        opts_params = f", {opts_str}"

    if required:
        if "default" not in opts:
            src = f"{name}: {pytype} = Field(...{opts_params})"
        else:
            src = f"{name}: {pytype} = Field({opts_str})"
    else:
        src = f"{name}: Optional[{pytype}] = Field({opts_str})"

    return src


def _generate_integer_field(name: str, required: bool, props: LexInteger) -> str:
    """
    Generate the source string for a pydantic field for an integer type.
    """
    opts = {}

    if props.minimum is not None:
        opts["ge"] = props.minimum
    if props.maximum is not None:
        opts["le"] = props.maximum
    if props.description is not None:
        opts["description"] = props.description
    if props.default is not None:
        opts["default"] = props.default

    assert props.const is None, "const not implemented"
    assert props.enum is None, "enum not implemented"

    s = _build_expanded("int", required, opts, name)

    return s


def _generate_bool_field(name: str, required: bool, props: LexBoolean) -> str:
    opts = {}

    if props.default is not None:
        opts["default"] = props.default

    if props.description is not None:
        opts["description"] = props.description

    if name == "validate":
        # Pydantic uses 'validate' don't shadow it'
        name = "validate_flag"
        opts["alias"] = "validate"

    # assert props.const is None, "const not implemented" TODO

    return _build_expanded("bool", required, opts, name)


class Symbol:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name


def _generate_str_field(
    ctx: GenCtx, name: str, required: bool, props: LexString
) -> str:
    opts = {}

    if props.default is not None:
        opts["default"] = props.default
    if props.description is not None:
        opts["description"] = props.description
    if props.max_length is not None:
        opts["max_length"] = props.max_length

    assert props.const is None, "const not implemented"
    assert props.enum is None, "enum not implemented"
    # assert props.min_graphemes is None, "min_graphemes not implemented"
    # assert props.max_graphemes is None, "max_graphemes not implemented"

    # I don't think these are enforced like an enum
    if props.known_values:
        opts["known_values"] = props.known_values

    if props.format:
        opts["pre"] = True
        validator = _add_imports_and_get_validator(ctx, props.format)
        opts["validator"] = validator

    return _build_expanded("str", required, opts, name)


def _add_imports_and_get_validator(ctx: GenCtx, fmt: str) -> Symbol:
    if fmt == "cid":
        ctx.imports["psychonaut.lexicon.formats"].add("validate_cid")
        return Symbol("validate_cid")
    elif fmt == "nsid":
        ctx.imports["psychonaut.lexicon.formats"].add("validate_nsid")
        return Symbol("validate_nsid")
    elif fmt == "did":
        ctx.imports["psychonaut.lexicon.formats"].add("validate_did")
        return Symbol("validate_did")
    elif fmt == "handle":
        ctx.imports["psychonaut.lexicon.formats"].add("validate_handle")
        return Symbol("validate_handle")
    elif fmt == "at-identifier":
        ctx.imports["psychonaut.lexicon.formats"].add("validate_at_identifier")
        return Symbol("validate_at_identifier")
    elif fmt == "at-uri":
        ctx.imports["psychonaut.lexicon.formats"].add("validate_at_uri")
        return Symbol("validate_at_uri")
    elif fmt == "datetime":
        ctx.imports["psychonaut.lexicon.formats"].add("validate_datetime")
        return Symbol("validate_datetime")
    elif fmt == "uri":
        ctx.imports["psychonaut.lexicon.formats"].add("validate_uri")
        return Symbol("validate_uri")
    else:
        assert False, f"Unknown format: {fmt}"


def _generate_array_field(
    ctx: GenCtx, name: str, required: bool, props: LexArray
) -> str:
    ctx.imports["typing"].add("List")
    ctx.imports["typing"].add("Optional")  # TODO: fix

    opts = {}

    if props.description is not None:
        opts["description"] = props.description

    t = Symbol(f"List[{_lex_type_to_pydantic_type.get(props.items.type, 'Any')}]")

    if props.min_length is not None:
        opts["min_items"] = props.min_length
    if props.max_length is not None:
        opts["max_items"] = props.max_length

    for unimp in ["const", "enum", "known_values", "min_graphemes", "max_graphemes"]:
        if hasattr(props.items, unimp) and getattr(props.items, unimp) is not None:
            assert False, f"{unimp} not implemented"

    if hasattr(props.items, "format") and props.items.format is not None:
        opts["pre"] = True
        underlying_validator = _add_imports_and_get_validator(ctx, props.items.format)
        ctx.imports["psychonaut.lexicon.formats"].add("validate_array")
        opts["validator"] = Symbol(f"validate_array({underlying_validator})")

    return _build_expanded(t, required, opts, name)


def _generate_generic_field(name: str, required: bool, props: Dict[str, Any]) -> str:
    t = _lex_type_to_pydantic_type.get(props.type, "Any")

    if required:
        return f"{name}: {t}"
    else:
        return f"{name}: Optional[{t}] = None"


def _generate_field(
    ctx: GenCtx, name: str, required: bool, props: Dict[str, Any]
) -> str:
    type_ = props.type

    if type_ == "array" and props.items.type == "ref":
        # print(f"Ref to any arr kludge: {ctx.input_path} {name}")
        # print(">>>>>>>>>>>>>>>>>>>>>>", ctx, name, props)
        type_ = "Any"

    if not required:
        ctx.imports["typing"].add("Optional")

    if type_ == "ref":
        return _generate_ref_field(ctx, name, required, props)
    if type_ == "integer":
        return _generate_integer_field(name, required, props)
    elif type_ == "string":
        return _generate_str_field(ctx, name, required, props)
    elif type_ == "boolean":
        return _generate_bool_field(name, required, props)
    elif type_ == "array":
        return _generate_array_field(ctx, name, required, props)
    else:
        print(f"Unknown type: {type_}")
        ctx.imports["typing"].add("Any")
        return _generate_generic_field(name, required, props)


def _generate_ref_field(
    ctx: GenCtx, name: str, required: bool, props: Dict[str, Any]
) -> str:
    ref_class = ""
    if "#" in props.ref:
        if props.ref.startswith("#"):
            ref_class = camel_to_class_name(props.ref[1:])
        elif len(parts := props.ref.split("#")) > 1:
            module = camel_to_snake(parts[0])
            ref_class = camel_to_class_name(parts[1])
            if module != ctx.common_xrpc_id:
                print("YYYYYY", props.ref, module, parts, ctx.common_xrpc_id)
                ctx.imports[MODULE_PREFIX + module].add(camel_to_class_name(parts[-1]))
    else:
        # Assume it's importing the main class at that path
        parts = props.ref.split(".")
        module = MODULE_PREFIX + camel_to_snake(props.ref)
        ref_class = camel_to_class_name(parts[-1])

        ctx.imports[module].add(camel_to_class_name(parts[-1]))

        print("YYY", props.ref, module, ref_class)

        ref_class = camel_to_class_name(props.ref.split(".")[-1])

    ref_class = camel_to_class_name(ref_class)
    if required:
        return f"{name}: {ref_class}"
    else:
        return f"{name}: Optional[{ref_class}] = None"


# def _generate_ref_field(ctx: GenCtx, name: str, required: bool, props: Dict[str, Any]) -> str:



from typing import Any, Optional
from psychonaut.api.lexicons.com.atproto.admin.defs import ActionView
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class GetModerationActionsResp(BaseModel):
    cursor: Optional[str] = Field(default=None)
    actions: Any


class GetModerationActionsReq(BaseModel):
    """
    List moderation actions related to a subject.
    """

    subject: Optional[str] = Field(default=None)
    limit: Optional[int] = Field(default=50, ge=1, le=100)
    cursor: Optional[str] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.admin.getModerationActions"

    async def do_xrpc(self, sess: Session) -> GetModerationActionsResp:
        resp = await sess.query(self)
        return GetModerationActionsResp(**resp)



from typing import Any
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_uri, validate_cid


class StrongRef(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: str = Field(..., pre=True, validator=validate_cid)



import pytest
from .nsid import NSID, InvalidNsidError, ensure_valid_nsid, ensure_valid_nsid_regex


def test_nsid_parsing():
    nsid = NSID.parse("com.example.foo")
    assert nsid.authority == "example.com"
    assert nsid.name == "foo"
    assert str(nsid) == "com.example.foo"

    nsid = NSID.parse("com.example.*")
    assert nsid.authority == "example.com"
    assert nsid.name == "*"
    assert str(nsid) == "com.example.*"

    nsid = NSID.parse("com.long-thing1.cool.fooBarBaz")
    assert nsid.authority == "cool.long-thing1.com"
    assert nsid.name == "fooBarBaz"
    assert str(nsid) == "com.long-thing1.cool.fooBarBaz"


def test_nsid_creation():
    nsid = NSID.create("example.com", "foo")
    assert nsid.authority == "example.com"
    assert nsid.name == "foo"
    assert str(nsid) == "com.example.foo"

    nsid = NSID.create("example.com", "*")
    assert nsid.authority == "example.com"
    assert nsid.name == "*"
    assert str(nsid) == "com.example.*"

    nsid = NSID.create("cool.long-thing1.com", "fooBarBaz")
    assert nsid.authority == "cool.long-thing1.com"
    assert nsid.name == "fooBarBaz"
    assert str(nsid) == "com.long-thing1.cool.fooBarBaz"


def expect_valid(nsid: str):
    ensure_valid_nsid(nsid)
    ensure_valid_nsid_regex(nsid)


def expect_invalid(nsid: str):
    with pytest.raises(InvalidNsidError):
        ensure_valid_nsid(nsid)
    with pytest.raises(InvalidNsidError):
        ensure_valid_nsid_regex(nsid)


def test_nsid_validation():
    expect_valid("com.example.foo")
    long_nsid = "com." + "o" * 63 + ".foo"
    expect_valid(long_nsid)

    too_long_nsid = "com." + "o" * 64 + ".foo"
    expect_invalid(too_long_nsid)

    long_end = "com.example." + "o" * 128
    expect_valid(long_end)

    too_long_end = "com.example." + "o" * 129
    expect_invalid(too_long_end)

    long_overall = "com." + "middle." * 50 + "foo"
    assert len(long_overall) == 357
    expect_valid(long_overall)

    too_long_overall = "com." + "middle." * 100 + "foo"
    assert len(too_long_overall) == 707
    expect_invalid(too_long_overall)

    expect_valid("a.b.c")
    expect_valid("a0.b1.c3")
    expect_valid("a-0.b-1.c-3")
    expect_valid("m.xn--masekowski-d0b.pl")
    expect_valid("one.two.three")

    expect_invalid("example.com")
    expect_invalid("com.example")
    expect_invalid("a.0.c")
    expect_invalid("a.")
    expect_invalid(".one.two.three")
    expect_invalid("one.two.three ")
    expect_invalid("one.two..three")
    expect_invalid("one .two.three")
    expect_invalid(" one.two.three")
    expect_invalid("com.exa💩ple.thing")
    expect_invalid("com.atproto.feed.p@st")
    expect_invalid("com.atproto.feed.p_st")
    expect_invalid("com.atproto.feed.p*st")
    expect_invalid("com.atproto.feed.po#t")
    expect_invalid("com.atproto.feed.p!ot")
    expect_invalid("com.example-.foo")


def test_nsid_onion():
    expect_valid("onion.expyuzz4wqqyqhjn.spec.getThing")
    expect_valid(
        "onion.g2zyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid.lex.deleteThing"
    )


def test_nsid_blocks_numeric():
    expect_invalid("org.4chan.lex.getThing")
    expect_invalid("cn.8.lex.stuff")
    expect_invalid(
        "onion.2gzyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid.lex.deleteThing"
    )



from typing import Any, Optional, List
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import (
    validate_at_identifier,
    validate_handle,
    validate_array,
    validate_did,
    validate_nsid,
)


class DescribeRepoResp(BaseModel):
    handle: str = Field(..., pre=True, validator=validate_handle)
    did: str = Field(..., pre=True, validator=validate_did)
    didDoc: Any
    collections: List[str] = Field(
        ..., pre=True, validator=validate_array(validate_nsid)
    )
    handleIsCorrect: bool = Field(...)


class DescribeRepoReq(BaseModel):
    """
    Get information about the repo, including the list of collections.
    """

    repo: str = Field(
        ...,
        description="The handle or DID of the repo.",
        pre=True,
        validator=validate_at_identifier,
    )

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.repo.describeRepo"

    async def do_xrpc(self, sess: Session) -> DescribeRepoResp:
        resp = await sess.query(self)
        return DescribeRepoResp(**resp)



from contextlib import contextmanager
from dotenv import load_dotenv


import json
import os
from pathlib import Path
from typing import Tuple


def load_creds(
    inject_dot_env: bool = True, allow_homedir_creds: bool = True
) -> Tuple[str, str]:
    """
    Load credentials from the environment or from ~/.bsky.json
    :param inject_dot_env: If true, injects the .env file into the environment.
        expects BSKY_USERNAME and BSKY_PASSWORD to be set
    :param allow_homedir_creds: If true, allows ~/.bsky.json to be used for credentials
    """
    if inject_dot_env:
        load_dotenv()

    username = os.getenv("BSKY_USERNAME", "")
    password = os.getenv("BSKY_PASSWORD", "")

    # attempt to lookup ~/.bsky.json for credentials
    if allow_homedir_creds and not username and not password:
        cred_path = Path.home() / ".psychonaut.json"
        if cred_path.exists():
            with cred_path.open("r") as fp:
                cred = json.load(fp)
                username = cred["username"]
                password = cred["password"]

    return username, password



@contextmanager
def write_homedir_creds(username: str, allow_overwrite: bool):
    """
    Write credentials to ~/.bsky.json

    :param username: The username to write
    :param allow_overwrite: If true, allows overwriting the file otherwise 
        raises FileExistsError
    """
    print("......")
    file_path = Path().home() / ".psychonaut.json"

    if file_path.exists() and not allow_overwrite:
        raise FileExistsError(f"File {file_path} exists")

    user_input = []
    def set_password(password: str):
        user_input.append(password)

    yield set_password

    # write the file
    file_path.write_text(
        json.dumps({
            "username": username,
            "password": user_input[0],
        })
    )



from typing import Any, Optional
from psychonaut.api.lexicons.com.atproto.admin.defs import RepoView
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class SearchReposResp(BaseModel):
    cursor: Optional[str] = Field(default=None)
    repos: Any


class SearchReposReq(BaseModel):
    """
    Find repositories based on a search term.
    """

    term: Optional[str] = Field(default=None)
    invitedBy: Optional[str] = Field(default=None)
    limit: Optional[int] = Field(default=50, ge=1, le=100)
    cursor: Optional[str] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.admin.searchRepos"

    async def do_xrpc(self, sess: Session) -> SearchReposResp:
        resp = await sess.query(self)
        return SearchReposResp(**resp)



from typing import Any, Optional
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import (
    validate_at_uri,
    validate_handle,
    validate_did,
    validate_datetime,
)
from psychonaut.api.lexicons.com.atproto.label.defs import Label


class ViewerState(BaseModel):
    """
    [none provided by spec]
    """

    muted: Optional[bool] = Field(default=None)
    blockedBy: Optional[bool] = Field(default=None)
    blocking: Optional[str] = Field(default=None, pre=True, validator=validate_at_uri)
    following: Optional[str] = Field(default=None, pre=True, validator=validate_at_uri)
    followedBy: Optional[str] = Field(default=None, pre=True, validator=validate_at_uri)


class ProfileViewBasic(BaseModel):
    """
    [none provided by spec]
    """

    did: str = Field(..., pre=True, validator=validate_did)
    handle: str = Field(..., pre=True, validator=validate_handle)
    displayName: Optional[str] = Field(default=None, max_length=640)
    avatar: Optional[str] = Field(default=None)
    viewer: Optional[ViewerState] = None
    labels: Optional[Any] = None


class ProfileView(BaseModel):
    """
    [none provided by spec]
    """

    did: str = Field(..., pre=True, validator=validate_did)
    handle: str = Field(..., pre=True, validator=validate_handle)
    displayName: Optional[str] = Field(default=None, max_length=640)
    description: Optional[str] = Field(default=None, max_length=2560)
    avatar: Optional[str] = Field(default=None)
    indexedAt: Optional[str] = Field(
        default=None, pre=True, validator=validate_datetime
    )
    viewer: Optional[ViewerState] = None
    labels: Optional[Any] = None


class ProfileViewDetailed(BaseModel):
    """
    [none provided by spec]
    """

    did: str = Field(..., pre=True, validator=validate_did)
    handle: str = Field(..., pre=True, validator=validate_handle)
    displayName: Optional[str] = Field(default=None, max_length=640)
    description: Optional[str] = Field(default=None, max_length=2560)
    avatar: Optional[str] = Field(default=None)
    banner: Optional[str] = Field(default=None)
    followersCount: Optional[int] = Field(default=None)
    followsCount: Optional[int] = Field(default=None)
    postsCount: Optional[int] = Field(default=None)
    indexedAt: Optional[str] = Field(
        default=None, pre=True, validator=validate_datetime
    )
    viewer: Optional[ViewerState] = None
    labels: Optional[Any] = None



from typing import Any, Optional
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_datetime


class GetUnreadCountResp(BaseModel):
    count: int = Field(...)


class GetUnreadCountReq(BaseModel):
    """
    [none provided by spec]
    """

    seenAt: Optional[str] = Field(default=None, pre=True, validator=validate_datetime)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.notification.getUnreadCount"

    async def do_xrpc(self, sess: Session) -> GetUnreadCountResp:
        resp = await sess.query(self)
        return GetUnreadCountResp(**resp)



from typing import Any, Optional
from psychonaut.api.lexicons.app.bsky.feed.defs import (
    ThreadViewPost,
    BlockedPost,
    NotFoundPost,
)
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_uri


class GetPostThreadResp(BaseModel):
    thread: Any


class GetPostThreadReq(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    depth: Optional[int] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.feed.getPostThread"

    async def do_xrpc(self, sess: Session) -> GetPostThreadResp:
        resp = await sess.query(self)
        return GetPostThreadResp(**resp)



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class NotifyOfUpdateReq(BaseModel):
    """
    Notify a crawling service of a recent update. Often when a long break
    between updates causes the connection with the crawling service to
    break.
    """

    hostname: str = Field(
        ..., description="Hostname of the service that is notifying of update."
    )

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.sync.notifyOfUpdate"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.query(self)



from typing import Any, Optional, List
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_uri, validate_did


class ByteSlice(BaseModel):
    """
    A text segment. Start is inclusive, end is exclusive. Indices are for
    utf8-encoded strings.
    """

    byteStart: int = Field(..., ge=0)
    byteEnd: int = Field(..., ge=0)


class Mention(BaseModel):
    """
    A facet feature for actor mentions.
    """

    did: str = Field(..., pre=True, validator=validate_did)


class Link(BaseModel):
    """
    A facet feature for links.
    """

    uri: str = Field(..., pre=True, validator=validate_uri)


class Facet(BaseModel):
    """
    [none provided by spec]
    """

    index: ByteSlice
    features: List[Any] = Field(...)



from typing import Any
from pydantic import BaseModel, Field


class Image(BaseModel):
    """
    [none provided by spec]
    """

    image: Any
    alt: str = Field(...)


class ViewImage(BaseModel):
    """
    [none provided by spec]
    """

    thumb: str = Field(...)
    fullsize: str = Field(...)
    alt: str = Field(...)


class Images(BaseModel):
    """
    [none provided by spec]
    """

    images: Any


class View(BaseModel):
    """
    [none provided by spec]
    """

    images: Any



import json
from pathlib import Path
from typing import Iterable
import jmespath


def collect_all_values(lexicon_dir: str, query: str, transform=lambda d: d) -> Iterable[str]:
    """
    Iterate over every .json file in the lexicon directory and collect all values
    """
    for lexicon_file in Path(lexicon_dir).glob("**/*.json"):
        with open(lexicon_file, "r") as f:
            lexicon = json.load(f)
            res = jmespath.search(query, lexicon)
            if res:
                res = transform(res)
                for value in res:
                    yield value



def find_doc_match(lexicon_dir: str, pred_f=lambda d: bool):
    for lexicon_file in Path(lexicon_dir).glob("**/*.json"):
        with open(lexicon_file, "r") as f:
            src = f.read()
            lexicon = json.loads(src)

            if pred_f(lexicon):
                print(lexicon_file)
                print(src)
                print()


from typing import Any, Optional
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_did, validate_handle


class GetSessionResp(BaseModel):
    handle: str = Field(..., pre=True, validator=validate_handle)
    did: str = Field(..., pre=True, validator=validate_did)
    email: Optional[str] = Field(default=None)


class GetSessionReq(BaseModel):
    """
    Get information about the current session.
    """

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.server.getSession"

    async def do_xrpc(self, sess: Session) -> GetSessionResp:
        resp = await sess.query(self)
        return GetSessionResp(**resp)



import asyncio
from datetime import datetime
from typing import Any, Dict, Optional
import aiohttp
from pydantic import BaseModel
from contextlib import asynccontextmanager


@asynccontextmanager
async def retires_on_timeout(max_retries: int = 3):

    for i in range(max_retries):
        try:
            yield
            return
        except asyncio.TimeoutError as e:
            print(f"TimeoutError {i}: {e}")
            continue

    raise aiohttp.ClientTimeoutError(f"TimeoutError after {max_retries} retries")

class Session:
    def __init__(self, factory, username: str, access_jwt: str, did: str):
        self.factory = factory
        self.username = username
        self.access_jwt = access_jwt
        self.did = did

    @property
    def http(self):
        return self.factory.http_session

    async def query(self, req, max_retries=3) -> Dict[Any, Any]:
        endpoint = self.factory.atp_host + f"/xrpc/{req.xrpc_id}"

        # TODO: when?
        headers = {"Authorization": f"Bearer {self.access_jwt}"}
        # I think this is true (for query methods?): TODO
        # if not req.xrpc_id.startswith("com.atproto."):
        #     headers = {"Authorization": f"Bearer {self.access_jwt}"}
        params = req.dict(exclude_none=True)
        #headers = None

        async with retires_on_timeout(max_retries=max_retries):
            async with self.http.get(endpoint, params=params, headers=headers) as resp:
                resp.raise_for_status()

                doc = await resp.json()

                if "error" in doc:
                    raise Exception(doc["error"])

                return doc

    async def _post_blob_kludge(self, data: bytes, content_type: str) -> Dict[Any, Any]:
        endpoint = self.factory.atp_host + "/xrpc/com.atproto.repo.uploadBlob"
        headers = {
            "Authorization": f"Bearer {self.access_jwt}",
            "Content-Type": content_type,
        }

        async with retires_on_timeout():
            async with self.http.post(
                endpoint, headers=headers, timeout=60, data=data
            ) as resp:
                resp.raise_for_status()

                doc = await resp.json()

                if "error" in doc:
                    raise Exception(doc["error"])

                return doc

    async def record(self, req) -> Dict[Any, Any]:
        endpoint = self.factory.atp_host + "/xrpc/com.atproto.repo.createRecord"
        headers = {"Authorization": f"Bearer {self.access_jwt}"}

        record = req.dict(exclude_none=True)
        if "createdAt" not in record:
            record["createdAt"] = datetime.utcnow().isoformat()
        if "$type" not in record:
            record["$type"] = req.xrpc_id

        req_data = {
            "collection": req.xrpc_id,
            "$type": req.xrpc_id,
            "repo": self.did,
            "record": record
        }

        async with retires_on_timeout():
            async with self.factory.http_session.post(
                endpoint, json=req_data, headers=headers, timeout=60
            ) as resp:
                resp.raise_for_status()

                doc = await resp.json()

                if "error" in doc:
                    raise Exception(doc["error"])

                return doc


class SessionFactory(BaseModel):
    atp_host: str
    http_session: aiohttp.ClientSession

    async def create(self, username: str, password: str):
        data = {"identifier": username, "password": password}

        endpoint = self.atp_host + "/xrpc/com.atproto.server.createSession"
        async with self.http_session.post(endpoint, json=data) as resp:
            doc = await resp.json()

        if "error" in doc:
            raise Exception(doc["error"])

        return Session(self, username, doc["accessJwt"], doc["did"])

    class Config:
        arbitrary_types_allowed = True



import re

from psychonaut.identifier.reserved import RESERVED_SUBDOMAINS


class InvalidHandleError(ValueError):
    pass


class ReservedHandleError(ValueError):
    pass


class UnsupportedDomainError(ValueError):
    pass


def ensure_valid_handle(handle: str) -> None:
    if not re.match(r"^[a-zA-Z0-9.-]*$", handle):
        raise InvalidHandleError(
            "Disallowed characters in handle (ASCII letters, digits, dashes, periods only)"
        )

    if len(handle) > 253:
        raise InvalidHandleError("Handle is too long (253 chars max)")

    labels = handle.split(".")
    if len(labels) < 2:
        raise InvalidHandleError("Handle domain needs at least two parts")

    for i, label in enumerate(labels):
        if len(label) < 1:
            raise InvalidHandleError("Handle parts cannot be empty")
        if len(label) > 63:
            raise InvalidHandleError("Handle part too long (max 63 chars)")
        if label.endswith("-") or label.startswith("-"):
            raise InvalidHandleError("Handle parts cannot start or end with hyphens")
        if i + 1 == len(labels) and not re.match(r"^[a-zA-Z]", label):
            raise InvalidHandleError(
                "Handle final component (TLD) must start with ASCII letter"
            )


def ensure_valid_handle_regex(handle: str) -> None:
    if not re.match(
        r"^([a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?\.)+[a-zA-Z]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?$",
        handle,
    ):
        raise InvalidHandleError("Handle didn't validate via regex")
    if len(handle) > 253:
        raise InvalidHandleError("Handle is too long (253 chars max)")


def normalize_handle(handle: str) -> str:
    return handle.lower()


def normalize_and_ensure_valid_handle(handle: str) -> str:
    normalized = normalize_handle(handle)
    ensure_valid_handle(normalized)
    return normalized


def is_valid_handle(handle: str) -> bool:
    try:
        ensure_valid_handle(handle)
    except InvalidHandleError:
        return False
    return True


def ensure_handle_service_constraints(
    handle: str, available_user_domains: list, reserved=None
) -> None:
    if reserved is None:
        reserved = RESERVED_SUBDOMAINS

    supported_domain = next(
        (domain for domain in available_user_domains if handle.endswith(domain)), None
    )

    if not supported_domain:
        raise UnsupportedDomainError("Not a supported handle domain")

    front = handle[: -len(supported_domain)]

    if "." in front:
        raise InvalidHandleError("Invalid characters in handle")

    if len(front) < 3:
        raise InvalidHandleError("Handle too short")

    if len(handle) > 30:
        raise InvalidHandleError("Handle too long")

    if front in reserved:
        raise ReservedHandleError("Reserved handle")


def fulfills_handle_service_constraints(
    handle: str, available_user_domains: list, reserved=None
) -> bool:
    try:
        ensure_handle_service_constraints(handle, available_user_domains, reserved)
    except (InvalidHandleError, ReservedHandleError, UnsupportedDomainError):
        return False
    return True



"""
See:

https://github.dev/bluesky-social/atproto/blob/main/packages/lexicon/src/validators/formats.ts
"""
import re


class InvalidDidError(ValueError):
    pass


def ensure_valid_did(did: str) -> None:
    # check that all chars are boring ASCII
    if not re.match(r"^[a-zA-Z0-9._:%-]*$", did):
        raise InvalidDidError(
            "Disallowed characters in DID (ASCII letters, digits, and a couple other characters only)"
        )

    parts = did.split(":")
    if len(parts) < 3:
        raise InvalidDidError(
            "DID requires prefix, method, and method-specific content"
        )

    if parts[0] != "did":
        raise InvalidDidError('DID requires "did:" prefix')

    if not re.match(r"^[a-z]+$", parts[1]):
        raise InvalidDidError("DID method must be lower-case letters")

    if did.endswith(":") or did.endswith("%"):
        raise InvalidDidError('DID can not end with ":" or "%"')

    if len(did) > 8 * 1024:
        raise InvalidDidError("DID is far too long")


def ensure_valid_did_regex(did: str) -> None:
    # simple regex to enforce most constraints via just regex and length
    if not re.match(r"^did:[a-z]+:[a-zA-Z0-9._:%-]*[a-zA-Z0-9._-]$", did):
        raise InvalidDidError("DID didn't validate via regex")

    if len(did) > 8 * 1024:
        raise InvalidDidError("DID is far too long")



import pytest
from .formats import validate_uri, validate_cid, validate_datetime, validate_nsid, validate_at_identifier


def test_validate_datetime():
    good = [
        "2022-12-12T00:50:36.809Z",
        "2022-12-12T00:50:36Z",
        "2022-12-12T00:50:36.8Z",
        "2022-12-12T00:50:36.80Z",
        "2022-12-12T00:50:36+00:00",
        "2022-12-12T00:50:36.8+00:00",
        "2022-12-11T19:50:36-05:00",
        "2022-12-11T19:50:36.8-05:00",
        "2022-12-11T19:50:36.80-05:00",
        "2022-12-11T19:50:36.809-05:00",
    ]

    for s in good:
        validate_datetime(s)

    with pytest.raises(ValueError):
        assert not validate_datetime("2022-12-12T00:50:36.809")


def test_validate_uri():
    examples = {
        "https://example.com": True,
        "https://example.com/with/path": True,
        "https://example.com/with/path?and=query": True,
        "at://bsky.social": True,
        "did:example:test": True,
        "com.example.uri": False,
    }

    for uri, expected in examples.items():
        if expected:
            validate_uri(uri)
        else:
            with pytest.raises(ValueError):
                validate_uri(uri)


def test_validate_cid():
    validate_cid("bafyreidfayvfuwqa7qlnopdjiqrxzs6blmoeu4rujcjtnci5beludirz2a")
    with pytest.raises(ValueError):
        validate_cid("abapsdofiuwrpoiasdfuaspdfoiu")


def test_validate_nsid():
    validate_nsid("com.atproto.test")

    with pytest.raises(ValueError):
        validate_nsid("com.atproto-.test")


def test_validate_at_identifier():
    examples = {
        'bsky.test': True,
        'did:plc:12345678abcdefghijklmnop': True,    
        'bad id': False,
        '-bad-.test': False,    
    }

    for uri, expected in examples.items():
        if expected:
            validate_at_identifier(uri)
        else:
            with pytest.raises(ValueError):
                validate_at_identifier(uri)



from typing import Any, Optional
from pydantic import BaseModel, Field
from psychonaut.api.lexicons.com.atproto.repo.strong_ref import StrongRef
from psychonaut.lexicon.formats import validate_datetime


class TextSlice(BaseModel):
    """
    Deprecated. Use app.bsky.richtext instead -- A text segment. Start is
    inclusive, end is exclusive. Indices are for utf16-encoded strings.
    """

    start: int = Field(..., ge=0)
    end: int = Field(..., ge=0)


class ReplyRef(BaseModel):
    """
    [none provided by spec]
    """

    root: StrongRef
    parent: StrongRef


class Entity(BaseModel):
    """
    Deprecated: use facets instead.
    """

    index: TextSlice
    type: str = Field(..., description="Expected values are 'mention' and 'link'.")
    value: str = Field(...)


class Post(BaseModel):
    """
    [none provided by spec]
    """

    text: str = Field(..., max_length=3000)
    entities: Optional[Any] = None
    facets: Optional[Any] = None
    reply: Optional[ReplyRef] = None
    embed: Optional[Any] = None
    createdAt: str = Field(..., pre=True, validator=validate_datetime)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.feed.post"



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_identifier


class UpdateAccountEmailReq(BaseModel):
    """
    Administrative action to update an account's email
    """

    account: str = Field(
        ...,
        description="The handle or DID of the repo.",
        pre=True,
        validator=validate_at_identifier,
    )
    email: str = Field(...)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.admin.updateAccountEmail"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.procedure(self)



from pathlib import Path
from pydantic import BaseModel, Field
from collections import OrderedDict, defaultdict
from typing import Dict, Optional, Set, Any
from psychonaut.lexicon.planner import LexiconDef


class GenCtx(BaseModel):
    imports: Dict[str, Set[str]] = Field(default_factory=lambda: defaultdict(set))
    fragments: OrderedDict[str, LexiconDef] = Field(default_factory=OrderedDict)
    verbose: bool = Field(default=False, repr=False)
    input_path: Optional[Path] = Field(default=None, repr=False)
    global_vars: Dict[str, Any] = Field(default_factory=dict, repr=False)

    def add_fragment(self, name: str, node: LexiconDef):
        assert name not in self.fragments, f"Fragment {name} already exists"
        self.fragments[name] = node

    @property
    def common_xrpc_id(self) -> Optional[str]:
        if not self.fragments:
            return None
        
        xrpc_ids = {
            v.node_id.split("#")[0]
            for v in self.fragments.values()
        }

        if len(xrpc_ids) == 1:
            return xrpc_ids.pop()

        raise ValueError(f"Multiple xrpc ids found: {xrpc_ids}")

    class Config:
        arbitrary_types_allowed = True


import json
from pathlib import Path
import sys
import click
import websockets
import time
from psychonaut.firehose.exponential_backoff import FirehoseExponentialBackoff
from psychonaut.firehose.io import (
    LengthDelimitedStreamSegmentWriter,
    convert_b64_to_length_prefixed_all,
    iter_length_prefixed_paths,
    noop_segment_writer,
    stream_to_stdout,
)

from psychonaut.firehose.serde import (
    read_jsonified_event,
    _json_encode_kludge,
)
from .group import cli
from .util import as_async, print_error_and_fail
import asyncio


@cli.command()
@click.argument("output_dir", type=click.Path(exists=False))
@click.option("--tee", is_flag=True, help="Write to stdout as well as file")
@click.option("--tee-first-block-only", is_flag=True, help="Only print the first block")
@as_async
async def repos_firehose_stream(output_dir: str, tee: bool, tee_first_block_only: bool):
    segment_writer = noop_segment_writer()

    if output_dir:
        segment_writer = LengthDelimitedStreamSegmentWriter(
            output_dir,
            finalize_segment=lambda path, bytes_written: print(
                f"Wrote {bytes_written} bytes to {path}"
            ),
        )

    backoff = FirehoseExponentialBackoff()

    with segment_writer as writer:
        while True:
            try:
                await _stream_run(
                    output_dir, tee, writer, tee_first_block_only
                )  # TODO: clean up
                backoff.reset()
            except KeyboardInterrupt:
                raise
            except Exception as e:
                raise e
                sleep_time = backoff.next_sleep_time()
                print(f"Error: {e}", file=sys.stderr)
                print(f"\tSleeping for {sleep_time} seconds", file=sys.stderr)
                await asyncio.sleep(sleep_time)


def _has_one_block(jsonified_event):
    return "blocks" in jsonified_event and len(jsonified_event["blocks"]) > 0


async def _stream_run(
    output_dir: str, tee: bool, writer, tee_first_block_only: bool = False
):
    callbacks = []
    if tee:

        def stdout_callback(msg, jsonified_event):
            if jsonified_event:
                if tee_first_block_only and _has_one_block(jsonified_event):
                    jsonified_event["blocks"] = [jsonified_event["blocks"][0]]
                    jsonified_event["block_cids"] = [jsonified_event["block_cids"][0]]
                print(json.dumps(_json_encode_kludge(jsonified_event)))

        callbacks.append(stdout_callback)

    if output_dir:
        callbacks.append(writer)

    if not tee and not output_dir:
        print_error_and_fail("Must specify either --tee or --output-dir")

    uri = "wss://bsky.social/xrpc/com.atproto.sync.subscribeRepos"
    async with websockets.connect(uri) as websocket:
        n = 0
        start = time.time()
        while True:
            # Read the websocket message.
            msg = await websocket.recv()

            # msg = await asyncio.wait_for(websocket.recv(), timeout=1)

            # Read the first block.
            #

            # TODO: clean this up its kinda just a mess & bandaid
            jsonified_event = read_jsonified_event(msg)

            for f in callbacks:
                f(msg, jsonified_event)

            if not tee and n % 100 == 0:
                per_second = n / (time.time() - start)
                print(f"Received {n} messages {per_second:.2f}/s")

            n += 1


@cli.command
@click.argument("length_prefixed_file", type=click.Path(exists=False))
def repos_firehose_replay(length_prefixed_file: str):
    length_prefixed_file = Path(length_prefixed_file)
    if not length_prefixed_file.exists():
        print_error_and_fail(f"File {length_prefixed_file} does not exist")

    if length_prefixed_file.is_file():
        stream_to_stdout(length_prefixed_file)

    elif length_prefixed_file.is_dir():
        for p in iter_length_prefixed_paths(length_prefixed_file):
            stream_to_stdout(p)


@cli.command()
@click.argument("input_dir", type=click.Path(exists=True))
@click.argument("output_dir", type=click.Path(exists=False))
@click.option("--verbose", is_flag=True, help="Print progress")
@click.option("--skip-empty", is_flag=True, help="Skip files with 0 bytes")
def b64_dir_to_length_prefixed(
    input_dir: str, output_dir: str, verbose: bool, skip_empty: bool
):
    input_dir = Path(input_dir)
    output_dir = Path(output_dir)
    convert_b64_to_length_prefixed_all(input_dir, output_dir, verbose)



from typing import Any, Optional
from psychonaut.api.lexicons.app.bsky.actor.defs import ProfileView
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class GetBlocksResp(BaseModel):
    cursor: Optional[str] = Field(default=None)
    blocks: Any


class GetBlocksReq(BaseModel):
    """
    Who is the requester's account blocking?
    """

    limit: Optional[int] = Field(default=50, ge=1, le=100)
    cursor: Optional[str] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.graph.getBlocks"

    async def do_xrpc(self, sess: Session) -> GetBlocksResp:
        resp = await sess.query(self)
        return GetBlocksResp(**resp)



from typing import Any, Optional, List
from psychonaut.api.lexicons.com.atproto.admin.defs import RepoRef
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_cid, validate_did, validate_array


class TakeModerationActionReq(BaseModel):
    """
    Take a moderation action on a repo.
    """

    action: str = Field(
        ...,
        known_values=[
            "com.atproto.admin.defs#takedown",
            "com.atproto.admin.defs#flag",
            "com.atproto.admin.defs#acknowledge",
        ],
    )
    subject: Any
    subjectBlobCids: Optional[List[str]] = Field(
        default=None, pre=True, validator=validate_array(validate_cid)
    )
    createLabelVals: Optional[List[str]] = Field(default=None)
    negateLabelVals: Optional[List[str]] = Field(default=None)
    reason: str = Field(...)
    createdBy: str = Field(..., pre=True, validator=validate_did)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.admin.takeModerationAction"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.procedure(self)



from .ipld import json_to_ipld, ipld_to_json, ipld_equals


def test_ipld():
    person = {
        "name": "Alice",
        "age": 30,
        "friend": {
            "/": {
                "$link": "bafyreih3j6mikj3ocg2reh4ox62uthv7g5rgl6yvbdg3qkbhaya5omazie"  
            }
        }
    }

    person_ipld = json_to_ipld(person)
    person_json = ipld_to_json(person_ipld)
    assert ipld_equals(person, person_json)


from typing import Any, Optional
from psychonaut.api.lexicons.app.bsky.actor.defs import ProfileViewBasic
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class SearchActorsTypeaheadResp(BaseModel):
    actors: Any


class SearchActorsTypeaheadReq(BaseModel):
    """
    Find actor suggestions for a search term.
    """

    term: Optional[str] = Field(default=None)
    limit: Optional[int] = Field(default=50, ge=1, le=100)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.actor.searchActorsTypeahead"

    async def do_xrpc(self, sess: Session) -> SearchActorsTypeaheadResp:
        resp = await sess.query(self)
        return SearchActorsTypeaheadResp(**resp)



from typing import Any, Optional
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_did, validate_handle


class CreateAccountResp(BaseModel):
    accessJwt: str = Field(...)
    refreshJwt: str = Field(...)
    handle: str = Field(..., pre=True, validator=validate_handle)
    did: str = Field(..., pre=True, validator=validate_did)


class CreateAccountReq(BaseModel):
    """
    Create an account.
    """

    email: str = Field(...)
    handle: str = Field(..., pre=True, validator=validate_handle)
    inviteCode: Optional[str] = Field(default=None)
    password: str = Field(...)
    recoveryKey: Optional[str] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.server.createAccount"

    async def do_xrpc(self, sess: Session) -> CreateAccountResp:
        resp = await sess.procedure(self)
        return CreateAccountResp(**resp)



import click
from psychonaut.client.credentials import write_homedir_creds
from .util import clean_handle, print_error_and_fail
from .group import cli


@cli.command()
@click.argument("handle")
@click.option("--allow-overwrite", is_flag=True, default=False)
def save_login(handle: str, allow_overwrite: bool):
    handle = clean_handle(handle)
    try:
        with write_homedir_creds(handle, allow_overwrite) as f:
            f(click.prompt(f"Enter password for {handle}", hide_input=True))
    except FileExistsError as e:
        print_error_and_fail(f"{e}: use --allow-overwrite to overwrite")



from typing import Any, Dict
import cbor2
import base64
import io
from multiformats import CID
from pydantic import BaseModel
from carbox.car import read_car, Block
import dag_cbor

from psychonaut.firehose.events import CommitEvt, HandleEvt


def read_event_pair(msg: bytes):
    with io.BytesIO(msg) as fp:
        header = cbor2.load(fp)
        event = dag_cbor.decode(fp.read())
    return header, event


class UnknownKludge(BaseModel):
    header: Dict[str, Any]
    event: Dict[str, Any]


def read_enriched_event(msg: bytes):
    header, event = read_event_pair(msg)

    match header["t"]:
        case "#commit":
            return CommitEvt.parse_obj(event)
        case "#handle":
            return HandleEvt.parse_obj(event)
        case _:
            return UnknownKludge(header=header, event=event)


def read_jsonified_event(msg: bytes):
    header, event = read_event_pair(msg)

    if "t" not in header or header["t"] != "#commit":
        return None

    # TODO: "t==#handle"

    cids, blocks = read_car(event["blocks"])

    event['block_cids'] = cids
    event["blocks"] = blocks

    return event

    # return {
    #     'ks': list(obj.keys()),
    #     'repo': obj['repo'],
    #     'block': blocks
    # }


def _json_encode_kludge(d: Dict[Any, Any]) -> Any:
    for k, v in d.items():
        if isinstance(v, CID):
            d[k] = str(v)
        elif isinstance(v, bytes):
            d[k] = base64.b64encode(v).decode()
        elif isinstance(v, BaseModel):
            _json_encode_kludge(v.dict())
        elif isinstance(v, dict):
            _json_encode_kludge(v)
        elif isinstance(v, list):
            for i, x in enumerate(v):
                if isinstance(x, CID):
                    v[i] = str(x)
                elif isinstance(x, dict):
                    _json_encode_kludge(x)
                elif isinstance(x, Block):
                    v[i] = [str(x.cid), _json_encode_kludge(x.decoded)]

    return d



import re
from urllib.parse import parse_qs, urlencode
from more_itertools import nth

ATP_URI_REGEX = re.compile(
    r"^(at:\/\/)?((?:did:[a-z0-9:%-]+)|(?:[a-z0-9][a-z0-9.:-]*))(\/[^?#\s]*)?(\?[^#\s]+)?(#[^\s]+)?$",
    re.IGNORECASE,
)
RELATIVE_REGEX = re.compile(r"^(\/[^?#\s]*)?(\?[^#\s]+)?(#[^\s]+)?$", re.IGNORECASE)


class AtUri:
    def __init__(self, uri, base=None):
        if base:
            parsed = parse(base)
            if not parsed:
                raise ValueError(f"Invalid at uri: {base}")
            relative_parsed = parse_relative(uri)
            if not relative_parsed:
                raise ValueError(f"Invalid path: {uri}")
            parsed.update(relative_parsed)
        else:
            parsed = parse(uri)
            if not parsed:
                raise ValueError(f"Invalid at uri: {uri}")

        self.hash = parsed["hash"]
        self.host = parsed["host"]
        self.pathname = parsed["pathname"]
        self.search_params = parsed["search_params"]

    @classmethod
    def make(cls, handle_or_did, collection=None, rkey=None):
        uri_str = handle_or_did
        if collection:
            uri_str += f"/{collection}"
        if rkey:
            uri_str += f"/{rkey}"
        return cls(uri_str)

    @property
    def protocol(self):
        return "at:"

    @property
    def origin(self):
        return f"at://{self.host}"

    @property
    def hostname(self):
        return self.host

    @hostname.setter
    def hostname(self, value):
        self.host = value

    @property
    def search(self):
        return urlencode(self.search_params, doseq=True)

    @search.setter
    def search(self, value):
        self.search_params = parse_qs(_no_leading_question_mark(value))

    @property
    def collection(self):
        return nth(self.pathname.split("/"), 1, default="")

    @collection.setter
    def collection(self, value):
        parts = self.pathname.split("/")
        parts[0] = value
        self.pathname = "/".join(parts)

    @property
    def rkey(self):
        return nth(self.pathname.split("/"), 2, default="")

    @rkey.setter
    def rkey(self, value):
        parts = self.pathname.split("/")
        if not parts[0]:
            parts[0] = "undefined"
        if len(parts) == 1:
            parts.append(value)
        else:
            parts[1] = value
        self.pathname = "/".join(parts)

    @property
    def href(self):
        return self.__str__()

    def __str__(self):
        path = self.pathname or "/"
        if not path.startswith("/"):
            path = f"/{path}"
        qs = urlencode(self.search_params, doseq=True)
        if qs and not qs.startswith("?"):
            qs = f"?{qs}"
        hash = self.hash
        if hash and not hash.startswith("#"):
            hash = f"#{hash}"
        return f"at://{self.host}{path}{qs}{hash}"

def _no_leading_question_mark(qs):
    if qs.startswith("?"):
        return qs[1:]
    return qs

def _parse_search_params(m: re.Match, group_idx: int):
    if m.group(group_idx):
        return parse_qs(_no_leading_question_mark(m.group(group_idx)))
    return {}

def parse(uri_str):
    match = ATP_URI_REGEX.match(uri_str)
    if match:
        return {
            "hash": match.group(5) or "",
            "host": match.group(2) or "",
            "pathname": match.group(3) or "",
            "search_params": _parse_search_params(match, 4),
        }
    return None


def parse_relative(uri_str):
    match = RELATIVE_REGEX.match(uri_str)
    if match:
        return {
            "hash": match.group(3) or "",
            "pathname": match.group(1) or "",
            "search_params": _parse_search_params(match, 2),
        }
    return None



from psychonaut.lexicon.ctx import GenCtx
from psychonaut.lexicon.types import LexBoolean, LexInteger, LexString, LexXrpcParameters
from .codegen import generate_pydantic_model
from psychonaut.util import load_test_fixture


class TestGeneratePydanticModel:
    def test_from_lex_xrpc_parameters(self, load_test_fixture):
        expected = load_test_fixture("pydantic_model_lex_xrpc_parameters.txt", False)

        ctx = GenCtx()

        lex_xrpc_parameters = LexXrpcParameters(
            description="A test request",
            required=["a_bool"],
            properties={
                "a_bool": LexBoolean(
                    description="A boolean",
                    default=True,
                    #const=False,  # TODO: const is not supported yet
                ),
                "an_int": LexInteger(
                    description="An integer",
                    default=42,
                    minimum=0,
                    maximum=100,
                    #const=42,  # TODO: const is not supported yet
                ),
                "a_string": LexString(
                    description="A string",
                    minLength=1,
                    maxLength=100,
                )
            }
        )

        src = generate_pydantic_model(ctx, "TestReq", "test 1", lex_xrpc_parameters)

        assert src == expected




from typing import Any, Optional
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import (
    validate_at_uri,
    validate_cid,
    validate_at_identifier,
    validate_nsid,
)


class CreateRecordResp(BaseModel):
    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: str = Field(..., pre=True, validator=validate_cid)


class CreateRecordReq(BaseModel):
    """
    Create a new record.
    """

    repo: str = Field(
        ...,
        description="The handle or DID of the repo.",
        pre=True,
        validator=validate_at_identifier,
    )
    collection: str = Field(
        ...,
        description="The NSID of the record collection.",
        pre=True,
        validator=validate_nsid,
    )
    rkey: Optional[str] = Field(default=None, description="The key of the record.")
    validate_flag: Optional[bool] = Field(
        alias="validate", default=True, description="Validate the record?"
    )
    record: Any
    swapCommit: Optional[str] = Field(
        default=None,
        description="Compare and swap with the previous commit by cid.",
        pre=True,
        validator=validate_cid,
    )

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.repo.createRecord"

    async def do_xrpc(self, sess: Session) -> CreateRecordResp:
        resp = await sess.procedure(self)
        return CreateRecordResp(**resp)



from typing import Any, Optional
from psychonaut.api.lexicons.com.atproto.server.defs import InviteCode
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class GetAccountInviteCodesResp(BaseModel):
    codes: Any


class GetAccountInviteCodesReq(BaseModel):
    """
    Get all invite codes for a given account
    """

    includeUsed: Optional[bool] = Field(default=True)
    createAvailable: Optional[bool] = Field(default=True)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.server.getAccountInviteCodes"

    async def do_xrpc(self, sess: Session) -> GetAccountInviteCodesResp:
        resp = await sess.query(self)
        return GetAccountInviteCodesResp(**resp)



# async context manager
from contextlib import asynccontextmanager
import os
import aiohttp
from psychonaut.api.session import Session, SessionFactory
from psychonaut.client.credentials import load_creds


@asynccontextmanager
async def get_simple_client_session(
    inject_dot_env: bool = True, allow_homedir_creds: bool = True
) -> Session:
    """
    Initialize a simple client session for use with the api.

    :param inject_dot_env: If true, injects the .env file into the environment.
        expects BSKY_USERNAME and BSKY_PASSWORD to be set
        TODO: the new password thing
    """
    username, password = load_creds(
        inject_dot_env=inject_dot_env, allow_homedir_creds=allow_homedir_creds
    )

    if not username or not password:
        raise ValueError(
            "BSKY_USERNAME and BSKY_PASSWORD must be set in the environment or .env file"
        )

    session_opts = dict(
        timeout=aiohttp.ClientTimeout(total=20),
    )

    async with aiohttp.ClientSession(**session_opts) as sess:
        factory = SessionFactory(
            atp_host=os.environ.get("ATP_HOST", "https://bsky.social"),
            http_session=sess,
        )
        yield await factory.create(username=username, password=password)


async def get_ipynb_client_session(
    inject_dot_env: bool = True, allow_homedir_creds: bool = True
) -> Session:
    username, password = load_creds(
        inject_dot_env=inject_dot_env, allow_homedir_creds=allow_homedir_creds
    )

    if not username or not password:
        raise ValueError(
            "BSKY_USERNAME and BSKY_PASSWORD must be set in the environment or .env file"
        )

    session_opts = dict(
        timeout=aiohttp.ClientTimeout(total=20),
    )

    http_sess = aiohttp.ClientSession(**session_opts)
    factory = SessionFactory(
        atp_host=os.environ.get("ATP_HOST", "https://bsky.social"),
        http_session=http_sess,
    )
    return await factory.create(username=username, password=password)






from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class ResetPasswordReq(BaseModel):
    """
    Reset a user account password using a token.
    """

    token: str = Field(...)
    password: str = Field(...)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.server.resetPassword"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.procedure(self)



import re
from multiformats import CID
from psychonaut.nsid.nsid import ensure_valid_nsid, InvalidNsidError
from psychonaut.identifier.did import ensure_valid_did
from psychonaut.identifier.handle import ensure_valid_handle
from psychonaut.uri.validation import ensure_valid_at_uri
from typing import Callable

# https://github.com/bluesky-social/atproto/blob/main/packages/lexicon/src/validators/formats.ts

URI_RE = re.compile(r"^\w+:(?:\/\/)?[^\s/][^\s]*$")
ISO8601_RE = re.compile(
    r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d{1,3})?(Z|[+-]\d{2}:\d{2})$"
)


def validate_datetime(s: str):
    """
    Returns True if the string is a valid RFC 3339 datetime.
    """
    try:
        if ISO8601_RE.match(s) is not None:
            return
    except ValueError:
        pass
    raise ValueError(f"Invalid datetime: {s}")


def validate_uri(s: str):
    try:
        if URI_RE.match(s) is not None:
            return
    except ValueError:
        pass
    raise ValueError(f"Invalid URI: {s}")


def validate_at_uri(s: str):
    try:
        ensure_valid_at_uri(s)
        return
    except ValueError as e:
        raise ValueError(f"Invalid AT URI: {s}: {e}")


def validate_array(f: Callable[[str], None]):
    return lambda s: [f(x) for x in s]


def validate_nsid(s: str):
    try:
        ensure_valid_nsid(s)
        return
    except (InvalidNsidError, ValueError) as e:
        raise ValueError(f"Invalid NSID: {s}: {e}")


def validate_cid(s: str):
    try:
        CID.decode(s)
        return
    except ValueError as e:
        raise ValueError(f"Invalid CID: {s}: {e}")
    except KeyError as e:
        raise ValueError(f"Invalid CID: {s}: {e}")


def validate_did(s: str):
    ensure_valid_did(s)


def validate_handle(s: str):
    ensure_valid_handle(s)


def validate_at_identifier(s: str):
    try:
        validate_did(s)
    except ValueError:
        validate_handle(s)


from typing import Any, Optional
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import (
    validate_cid,
    validate_uri,
    validate_did,
    validate_datetime,
)


class Label(BaseModel):
    """
    Metadata tag on an atproto resource (eg, repo or record)
    """

    src: str = Field(
        ...,
        description="DID of the actor who created this label",
        pre=True,
        validator=validate_did,
    )
    uri: str = Field(
        ...,
        description="AT URI of the record, repository (account), or other resource which this label applies to",
        pre=True,
        validator=validate_uri,
    )
    cid: Optional[str] = Field(
        default=None,
        description="optionally, CID specifying the specific version of 'uri' resource this label applies to",
        pre=True,
        validator=validate_cid,
    )
    val: str = Field(
        ...,
        description="the short string name of the value or type of this label",
        max_length=128,
    )
    neg: Optional[bool] = Field(
        default=None,
        description="if true, this is a negation label, overwriting a previous label",
    )
    cts: str = Field(
        ...,
        description="timestamp when this label was created",
        pre=True,
        validator=validate_datetime,
    )



import pytest

from .handle import (
    ensure_valid_handle,
    ensure_valid_handle_regex,
    ensure_handle_service_constraints,
    normalize_and_ensure_valid_handle,
    InvalidHandleError,
)


def expect_valid(h: str):
    ensure_valid_handle(h)
    ensure_valid_handle_regex(h)

def expect_invalid(h: str):
    with pytest.raises(InvalidHandleError):
        ensure_valid_handle(h)
    with pytest.raises(InvalidHandleError):
        ensure_valid_handle_regex(h)


class TestHandleValidation:

    def test_valid_handles(self):
        long_handle = 'shoooort' + ('.loooooooooooooooooooooooooong' * 8) + '.test'  
        assert len(long_handle) == 253

        valid_handles = [
            'A.ISI.EDU',
            'XX.LCS.MIT.EDU',
            'SRI-NIC.ARPA',
            'john.test',
            'jan.test',
            'a234567890123456789.test',
            'john2.test',
            'john-john.test',
            'john.bsky.app',
            'jo.hn',
            'a.co',
            'a.org',
            'joh.n',
            'j0.h0',
            long_handle,
            'jaymome-johnber123456.test',
            'jay.mome-johnber123456.test',
            'john.test.bsky.app',
            'laptop.local',
            'laptop.arpa',
            'john.t',
        ]
        for handle in valid_handles:
            expect_valid(handle)


    def test_puny_handles(self):
        valid_punycode_handles = ['xn--ls8h.test', 'xn--bcher-kva.tld']
        for handle in valid_punycode_handles:
            expect_valid(handle)

    def test_throws_on_invalid_handles(self):
        expect_invalid('did:thing.test')
        expect_invalid('did:thing')
        expect_invalid('john-.test')
        expect_invalid('john.0')
        expect_invalid('john.-')
        expect_invalid('short.' + 'o' * 64 + '.test')
        expect_invalid('short' + '.loooooooooooooooooooooooong' * 10 + '.test')
        long_handle = 'shooooort' + '.loooooooooooooooooooooooooong' * 8 + '.test'
        assert len(long_handle) == 254
        expect_invalid(long_handle)
        expect_invalid('xn--bcher-.tld')
        expect_invalid('john..test')
        expect_invalid('jo_hn.test')
        expect_invalid('-john.test')
        expect_invalid('.john.test')
        expect_invalid('jo!hn.test')
        expect_invalid('jo%hn.test')
        expect_invalid('jo&hn.test')
        expect_invalid('jo@hn.test')
        expect_invalid('jo*hn.test')
        expect_invalid('jo|hn.test')
        expect_invalid('jo:hn.test')
        expect_invalid('jo/hn.test')
        expect_invalid('john💩.test')
        expect_invalid('bücher.test')
        expect_invalid('john .test')
        expect_invalid('john.test.')
        expect_invalid('john')
        expect_invalid('john.')
        expect_invalid('.john')
        expect_invalid('john.test.')
        expect_invalid('.john.test')
        expect_invalid(' john.test')
        expect_invalid('john.test ')
        expect_invalid('joh-.test')
        expect_invalid('john.-est')
        expect_invalid('john.tes-')

    def test_allows_onion_tor_handles(self):
        expect_valid('expyuzz4wqqyqhjn.onion')
        expect_valid('friend.expyuzz4wqqyqhjn.onion')
        expect_valid('g2zyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid.onion')
        expect_valid('friend.g2zyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid.onion')
        expect_valid('friend.g2zyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid.onion')
        expect_valid('2gzyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid.onion')
        expect_valid('friend.2gzyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid.onion')

    def test_correctly_validates_corner_cases_modern_vs_old_rfc(self):
        expect_valid('12345.test')
        expect_valid('8.cn')
        expect_valid('4chan.org')
        expect_valid('4chan.o-g')
        expect_valid('blah.4chan.org')
        expect_valid('thing.a01')
        expect_valid('120.0.0.1.com')
        expect_valid('0john.test')
        expect_valid('9sta--ck.com')
        expect_valid('99stack.com')
        expect_valid('0ohn.test')
        expect_valid('john.t--t')
        expect_valid('thing.0aa.thing')

        expect_invalid('cn.8')
        expect_invalid('thing.0aa')
        expect_invalid('thing.0aa')

    def test_does_not_allow_ip_addresses_as_handles(self):
        expect_invalid('127.0.0.1')
        expect_invalid('192.168.0.142')
        expect_invalid('fe80::7325:8a97:c100:94b')
        expect_invalid('2600:3c03::f03c:9100:feb0:af1f')

    def test_is_consistent_with_examples_from_stackoverflow(self):
        ok_stackoverflow = [
            'stack.com',
            'sta-ck.com',
            'sta---ck.com',
            'sta--ck9.com',
            'stack99.com',
            'sta99ck.com',
            'google.com.uk',
            'google.co.in',
            'google.com',
            'maselkowski.pl',
            'm.maselkowski.pl',
            'xn--masekowski-d0b.pl',
            'xn--fiqa61au8b7zsevnm8ak20mc4a87e.xn--fiqs8s',
            'xn--stackoverflow.com',
            'stackoverflow.xn--com',
            'stackoverflow.co.uk',
            'xn--masekowski-d0b.pl',
            'xn--fiqa61au8b7zsevnm8ak20mc4a87e.xn--fiqs8s',
        ]

        for domain in ok_stackoverflow:
            expect_valid(domain)

        bad_stackoverflow = [
            '-notvalid.at-all',
            '-thing.com',
            'www.masełkowski.pl.com',
        ]

        for domain in bad_stackoverflow:
            expect_invalid(domain)


class TestServiceConstraintsAndNormalization:
    domains = ['.bsky.app', '.test']

    def test_throw_on_handles_that_violate_service_constraints(self):
        def expect_throw(handle, err):
            with pytest.raises(Exception, match=err):
                ensure_handle_service_constraints(handle, self.domains)

        expect_throw('john.bsky.io', 'Not a supported handle domain')
        expect_throw('john.com', 'Not a supported handle domain')
        expect_throw('j.test', 'Handle too short')
        expect_throw('uk.test', 'Handle too short')
        expect_throw('john.test.bsky.app', 'Invalid characters in handle')
        expect_throw('about.test', 'Reserved handle')
        expect_throw('atp.test', 'Reserved handle')
        expect_throw('barackobama.test', 'Reserved handle')


    def test_normalizes_handles(self):
        normalized = normalize_and_ensure_valid_handle('JoHn.TeST')
        assert normalized == 'john.test'


    def test_throws_on_invalid_normalized_handles(self):
        with pytest.raises(InvalidHandleError):
            normalize_and_ensure_valid_handle('JoH!n.TeST')
